@@ -1,24 +1,84 @@
-from typing import Any, Tuple, Dict, List, Optional
-from itertools import count
+from __future__ import annotations
+from typing import Any
 import copy
 import logging
 
 import ailment.expression
 import networkx
 
-from ailment import Block
+from ailment import Block, AILBlockWalker
 from ailment.statement import Jump, ConditionalJump, Assignment, Return, Label
-from ailment.expression import Const
+from ailment.expression import Const, Phi, VirtualVariable
 
-from ..condition_processor import ConditionProcessor, EmptyBlockNotice
-from ..graph_region import GraphRegion
-from ..utils import remove_labels, to_ail_supergraph, calls_in_graph
-from ..structuring.structurer_nodes import MultiNode, ConditionNode
-from ..region_identifier import RegionIdentifier
+from angr.utils.ail import is_phi_assignment
+from angr.analyses.decompiler.condition_processor import ConditionProcessor, EmptyBlockNotice
+from angr.analyses.decompiler.graph_region import GraphRegion
+from angr.analyses.decompiler.utils import remove_labels, to_ail_supergraph, calls_in_graph
+from angr.analyses.decompiler.structuring.structurer_nodes import MultiNode, ConditionNode
+from angr.analyses.decompiler.region_identifier import RegionIdentifier
 
 _l = logging.getLogger(name=__name__)
 
 
+class FreshVirtualVariableRewriter(AILBlockWalker):
+    """
+    Helper class to rewrite virtual variables so that they will use fresh virtual variables.
+    """
+
+    def __init__(self, vvar_id_start: int, vvar_mapping: dict[int, int]):
+        super().__init__()
+        self.vvar_idx = vvar_id_start
+        self.vvar_mapping = vvar_mapping
+        self.new_block = None
+
+    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Block | None):
+        new_stmt = super()._handle_Assignment(stmt_idx, stmt, block)
+        dst = new_stmt.dst if new_stmt is not None else stmt.dst
+        src = new_stmt.src if new_stmt is not None else stmt.src
+        if isinstance(dst, VirtualVariable):
+            self.vvar_mapping[dst.varid] = self.vvar_idx
+            self.vvar_idx += 1
+
+            dst = VirtualVariable(
+                dst.idx,
+                self.vvar_mapping[dst.varid],
+                dst.bits,
+                dst.category,
+                dst.oident,
+                variable=dst.variable,
+                variable_offset=dst.variable_offset,
+                **dst.tags,
+            )
+
+            return Assignment(stmt.idx, dst, src, **stmt.tags)
+
+        return new_stmt
+
+    def _handle_VirtualVariable(  # type:ignore
+        self, expr_idx: int, expr: VirtualVariable, stmt_idx: int, stmt, block: Block | None
+    ) -> VirtualVariable | None:
+        if expr.varid in self.vvar_mapping:
+            return VirtualVariable(
+                expr.idx,
+                self.vvar_mapping[expr.varid],
+                expr.bits,
+                expr.category,
+                expr.oident,
+                variable=expr.variable,
+                variable_offset=expr.variable_offset,
+                **expr.tags,
+            )
+        return None
+
+    def _handle_stmt(self, stmt_idx: int, stmt, block: Block):  # type:ignore
+        r = super()._handle_stmt(stmt_idx, stmt, block)
+        if r is not None:
+            # replace the original statement
+            if self.new_block is None:
+                self.new_block = block.copy()
+            self.new_block.statements[stmt_idx] = r
+
+
 class ReturnDuplicatorBase:
     """
     The base class for implementing Return Duplication as described in the SAILR paper.
@@ -29,32 +89,45 @@ class ReturnDuplicatorBase:
     def __init__(
         self,
         func,
-        node_idx_start: int = 0,
+        *,
+        vvar_id_start: int,
         max_calls_in_regions: int = 2,
         minimize_copies_for_regions: bool = True,
-        ri: Optional[RegionIdentifier] = None,
-        **kwargs,
+        ri: RegionIdentifier | None = None,
+        scratch: dict[str, Any] | None = None,
+        max_func_blocks: int = 1500,
     ):
-        self.node_idx = count(start=node_idx_start)
         self._max_calls_in_region = max_calls_in_regions
         self._minimize_copies_for_regions = minimize_copies_for_regions
+        self._supergraph = None
 
         # this should also be set by the optimization passes initer
+        self.scratch = scratch if scratch is not None else {}
         self._func = func
-        self._ri: Optional[RegionIdentifier] = ri
+        self._ri: RegionIdentifier | None = ri
+        self.vvar_id_start = vvar_id_start
+        self._max_func_blocks = max_func_blocks
+
+    def next_node_idx(self) -> int:
+        node_idx = self.scratch.get("returndup_node_idx", 0) + 1
+        self.scratch["returndup_node_idx"] = node_idx
+        return node_idx
 
     #
     # must implement these methods
     #
 
     def _should_duplicate_dst(self, src, dst, graph, dst_is_const_ret=False) -> bool:
-        raise NotImplementedError()
+        raise NotImplementedError
 
     #
     # main analysis
     #
 
     def _check(self):
+        # is this function too large?
+        if len(self._func.block_addrs_set) > self._max_func_blocks:
+            return False, None
         # does this function have end points?
         return bool(self._func.endpoints), None
 
@@ -71,24 +144,39 @@ class ReturnDuplicatorBase:
             # for connected in_edges that form a region
             endnode_regions = self._copy_connected_edge_components(endnode_regions, graph)
 
+        # refresh the supergraph
+        self._supergraph = to_ail_supergraph(graph)
         for region_head, (in_edges, region) in endnode_regions.items():
             is_single_const_ret_region = self._is_simple_return_graph(region)
+            dup_pred_nodes = []
+            # duplicate the entire region if at least (N-2) in-edges for the region head is deemed should be duplicated.
+            # otherwise we only duplicate the edges that should be duplicated
             for in_edge in in_edges:
                 pred_node = in_edge[0]
                 if self._should_duplicate_dst(
                     pred_node, region_head, graph, dst_is_const_ret=is_single_const_ret_region
                 ):
+                    dup_pred_nodes.append(pred_node)
+
+            dup_count = len(dup_pred_nodes)
+            dup_all = dup_count >= len(in_edges) - 2 > 0
+            if dup_all:
+                for pred_node in sorted((in_edge[0] for in_edge in in_edges), key=lambda x: x.addr):
                     # every eligible pred gets a new region copy
                     self._copy_region([pred_node], region_head, region, graph)
+                    graph_changed = True
+            else:
+                for pred_node in dup_pred_nodes:
+                    self._copy_region([pred_node], region_head, region, graph)
+                    graph_changed = True
 
             if region_head in graph and graph.in_degree(region_head) == 0:
                 graph.remove_nodes_from(region)
-
-            graph_changed = True
+                graph_changed = True
 
         return graph_changed
 
-    def _find_endnode_regions(self, graph) -> Dict[Any, Tuple[List[Tuple[Any, Any]], networkx.DiGraph]]:
+    def _find_endnode_regions(self, graph) -> dict[Any, tuple[list[tuple[Any, Any]], networkx.DiGraph]]:
         """
         Find all the regions that contain a node with no successors. These are the "end nodes" of the graph.
         """
@@ -97,7 +185,7 @@ class ReturnDuplicatorBase:
         # to_update is keyed by the region head.
         # this is because different end nodes may lead to the same region head: consider the case of the typical "fork"
         # region where stack canary is checked in x86-64 binaries.
-        end_node_regions: Dict[Any, Tuple[List[Tuple[Any, Any]], networkx.DiGraph]] = {}
+        end_node_regions: dict[Any, tuple[list[tuple[Any, Any]], networkx.DiGraph]] = {}
 
         for end_node in endnodes:
             in_edges = list(graph.in_edges(end_node))
@@ -139,17 +227,23 @@ class ReturnDuplicatorBase:
 
         return end_node_regions
 
-    def _copy_region(self, pred_nodes, region_head, region, graph):
+    def _copy_region(self, pred_nodes: list[Block], region_head, region, graph):
         # copy the entire return region
-        copies = {}
-        queue = [(pred_node, region_head) for pred_node in pred_nodes]
+        copies: dict[Block, Block] = {}
+        queue: list[tuple[Block, Block]] = [(pred_node, region_head) for pred_node in pred_nodes]
+        vvar_mapping: dict[int, int] = {}
         while queue:
             pred, node = queue.pop(0)
             if node in copies:
                 node_copy = copies[node]
             else:
-                node_copy = copy.deepcopy(node)
-                node_copy.idx = next(self.node_idx)
+                if node is region_head:
+                    node_copy = self._copy_node_and_update_phi_variables(node, pred)
+                else:
+                    node_copy = copy.deepcopy(node)
+                node_copy = self._use_fresh_virtual_variables(node_copy, vvar_mapping)
+                node_copy.idx = self.next_node_idx()
+                self._fix_copied_node_labels(node_copy)
                 copies[node] = node_copy
 
             # modify Jump.target_idx and ConditionalJump.{true,false}_target_idx accordingly
@@ -158,12 +252,33 @@ class ReturnDuplicatorBase:
                 last_stmt = ConditionProcessor.get_last_statement(pred)
                 if isinstance(last_stmt, Jump):
                     if isinstance(last_stmt.target, Const) and last_stmt.target.value == node_copy.addr:
-                        last_stmt.target_idx = node_copy.idx
+                        updated_last_stmt = Jump(
+                            last_stmt.idx, last_stmt.target, target_idx=node_copy.idx, **last_stmt.tags
+                        )
+                        pred.statements[-1] = updated_last_stmt
                 elif isinstance(last_stmt, ConditionalJump):
                     if isinstance(last_stmt.true_target, Const) and last_stmt.true_target.value == node_copy.addr:
-                        last_stmt.true_target_idx = node_copy.idx
+                        updated_last_stmt = ConditionalJump(
+                            last_stmt.idx,
+                            last_stmt.condition,
+                            last_stmt.true_target,
+                            last_stmt.false_target,
+                            true_target_idx=node_copy.idx,
+                            false_target_idx=last_stmt.false_target_idx,
+                            **last_stmt.tags,
+                        )
+                        pred.statements[-1] = updated_last_stmt
                     elif isinstance(last_stmt.false_target, Const) and last_stmt.false_target.value == node_copy.addr:
-                        last_stmt.false_target_idx = node_copy.idx
+                        updated_last_stmt = ConditionalJump(
+                            last_stmt.idx,
+                            last_stmt.condition,
+                            last_stmt.true_target,
+                            last_stmt.false_target,
+                            true_target_idx=last_stmt.true_target_idx,
+                            false_target_idx=node_copy.idx,
+                            **last_stmt.tags,
+                        )
+                        pred.statements[-1] = updated_last_stmt
             except EmptyBlockNotice:
                 pass
 
@@ -173,11 +288,58 @@ class ReturnDuplicatorBase:
         for pred_node in pred_nodes:
             # delete the old edge to the return node
             graph.remove_edge(pred_node, region_head)
+            # update phi variables at the beginning of region_head
+            self._update_phi_variables_after_removing_predecessor(region_head, pred_node)
+
+    @staticmethod
+    def _copy_node_and_update_phi_variables(node: Block, pred: Block) -> Block:
+        stmts = []
+        for stmt in node.statements:
+            if isinstance(stmt, Assignment) and isinstance(stmt.src, Phi) and isinstance(stmt.dst, VirtualVariable):
+                # pick the variable from the correct source
+                vvar_src = next(iter(vvar for src, vvar in stmt.src.src_and_vvars if src == (pred.addr, pred.idx)), ...)
+                if vvar_src is ...:
+                    _l.warning(
+                        "Cannot find the virtual variable in Phi expression %r that corresponds with node %#x-%s.",
+                        stmt.src,
+                        pred.addr,
+                        pred.idx,
+                    )
+                elif vvar_src is None:
+                    # not used in this branch. drop this statement
+                    continue
+                else:
+                    new_stmt = Assignment(stmt.idx, stmt.dst, vvar_src, **stmt.tags)
+                    stmts.append(new_stmt)
+                    continue
+            stmts.append(stmt)
+
+        return Block(node.addr, node.original_size, statements=stmts, idx=node.idx)
+
+    @staticmethod
+    def _update_phi_variables_after_removing_predecessor(node: Block, pred: Block) -> None:
+        for idx in range(len(node.statements)):  # pylint:disable=consider-using-enumerate
+            stmt = node.statements[idx]
+            if isinstance(stmt, Assignment) and isinstance(stmt.src, Phi) and isinstance(stmt.dst, VirtualVariable):
+                # remove the variable from the specified source
+                new_src_and_vvars = [
+                    (src, vvar) for src, vvar in stmt.src.src_and_vvars if src != (pred.addr, pred.idx)
+                ]
+                new_phi = Phi(stmt.src.idx, stmt.src.bits, new_src_and_vvars, **stmt.src.tags)
+                node.statements[idx] = Assignment(stmt.idx, stmt.dst, new_phi, **stmt.tags)
+
+    def _use_fresh_virtual_variables(self, node: Block, vvar_map: dict[int, int]) -> Block:
+        rewriter = FreshVirtualVariableRewriter(self.vvar_id_start, vvar_map)
+        rewriter.walk(node)
+        self.vvar_id_start = rewriter.vvar_idx + 1
+        return rewriter.new_block if rewriter.new_block is not None else node
 
     def _copy_connected_edge_components(
-        self, endnode_regions: Dict[Any, Tuple[List[Tuple[Any, Any]], networkx.DiGraph]], graph: networkx.DiGraph
+        self, endnode_regions: dict[Any, tuple[list[tuple[Any, Any]], networkx.DiGraph]], graph: networkx.DiGraph
     ):
         updated_regions = endnode_regions.copy()
+        assert self._ri is not None
+        assert isinstance(self._ri.region, GraphRegion)
         all_region_block_addrs = list(self._find_block_sets_in_all_regions(self._ri.region).values())
         for region_head, (in_edges, region) in endnode_regions.items():
             is_single_const_ret_region = self._is_simple_return_graph(region)
@@ -247,7 +409,7 @@ class ReturnDuplicatorBase:
             return False
 
         # check if the graph is a single successor chain
-        if not all(labeless_graph.out_degree(n) <= 1 for n in nodes):
+        if not all(labeless_graph.out_degree[n] <= 1 for n in nodes):
             return False
 
         # collect the statements from the top node, make sure one exists
@@ -272,7 +434,7 @@ class ReturnDuplicatorBase:
                 return False
 
         # gather all assignments
-        assignments = [s for s in stmts if isinstance(s, Assignment)]
+        assignments = [s for s in stmts if isinstance(s, Assignment) and not is_phi_assignment(s)]
         has_assign = len(assignments) > 0
         if len(assignments) > max_assigns:
             return False
@@ -289,7 +451,15 @@ class ReturnDuplicatorBase:
         if ret_exprs and len(ret_exprs) > 1:
             return False
 
-        ret_expr = ret_exprs[0] if ret_exprs and len(ret_exprs) == 1 else None
+        if not ret_exprs:
+            # a simple return statement that does not carry any value or variable to return
+            return True
+
+        ret_expr = ReturnDuplicatorBase.unwrap_conv(ret_exprs[0])
+        # check if ret_expr is a virtual variable or not
+        if not isinstance(ret_expr, (VirtualVariable, Const)):
+            return False
+
         # stop early if there are no assignments at all and just jumps and rets, or a const ret
         if not has_assign:
             return True
@@ -301,9 +471,11 @@ class ReturnDuplicatorBase:
         # assignments to registers from the stack are valid, since cases of these assignments
         # pop up across optimized binaries
         elif (
-            isinstance(assign.dst, ailment.expression.Register)
-            and isinstance(assign.src, ailment.expression.Load)
-            and isinstance(assign.src.addr, ailment.expression.StackBaseOffset)
+            isinstance(assign.dst, ailment.expression.VirtualVariable)
+            and assign.dst.was_reg
+            and isinstance(assign.src, ailment.expression.VirtualVariable)
+            and assign.src.was_stack
+            and isinstance(assign.src.stack_offset, int)
         ):
             valid_assignment = True
         else:
@@ -312,7 +484,7 @@ class ReturnDuplicatorBase:
         return valid_assignment
 
     @staticmethod
-    def _single_entry_region(graph, end_node) -> Tuple[networkx.DiGraph, Any]:
+    def _single_entry_region(graph, end_node) -> tuple[networkx.DiGraph, Any]:
         """
         Back track on the graph from `end_node` and find the longest chain of nodes where each node has only one
         predecessor and one successor (the second-to-last node may have two successors to account for the typical
@@ -384,7 +556,7 @@ class ReturnDuplicatorBase:
         return region, region_head
 
     @staticmethod
-    def _is_indirect_jump_ailblock(block: "Block") -> bool:
+    def _is_indirect_jump_ailblock(block: Block) -> bool:
         if block.statements and isinstance(block.statements[-1], Jump):
             last_stmt = block.statements[-1]
             if not isinstance(last_stmt.target, Const):
@@ -412,9 +584,9 @@ class ReturnDuplicatorBase:
         def _unpack_block_type_to_addrs(node):
             if isinstance(node, Block):
                 return {node.addr}
-            elif isinstance(node, MultiNode):
+            if isinstance(node, MultiNode):
                 return {n.addr for n in node.nodes}
-            elif isinstance(node, ConditionNode):
+            if isinstance(node, ConditionNode):
                 return _unpack_block_type_to_addrs(node.true_node) | _unpack_block_type_to_addrs(node.false_node)
             return set()
 
@@ -446,3 +618,23 @@ class ReturnDuplicatorBase:
         all_region_block_sets = {}
         _unpack_every_region(top_region, all_region_block_sets)
         return all_region_block_sets
+
+    @staticmethod
+    def _fix_copied_node_labels(block: Block):
+        for i in range(len(block.statements)):  # pylint:disable=consider-using-enumerate
+            stmt = block.statements[i]
+            if isinstance(stmt, Label):
+                # fix the default name by suffixing it with the new block ID
+                new_name = stmt.name if stmt.name else f"Label_{stmt.ins_addr:x}"
+                if stmt.block_idx is not None:
+                    suffix = f"__{stmt.block_idx}"
+                    new_name = new_name.removesuffix(suffix)
+                else:
+                    new_name = stmt.name
+                new_name += f"__{block.idx}"
+
+                block.statements[i] = Label(stmt.idx, new_name, stmt.ins_addr, block_idx=block.idx, **stmt.tags)
+
+    @staticmethod
+    def unwrap_conv(expr):
+        return ReturnDuplicatorBase.unwrap_conv(expr.operand) if isinstance(expr, ailment.expression.Convert) else expr
