@@ -1,23 +1,23 @@
 # pylint:disable=arguments-renamed,global-statement
+from __future__ import annotations
 import copy
 import os
 import logging
 import inspect
 from collections import defaultdict
-from typing import Optional, Dict, Type, List, TYPE_CHECKING
-
-import itanium_demangler
+from typing import TYPE_CHECKING
 
+import pydemumble
 import archinfo
 
-from ...errors import AngrMissingTypeError
-from ...sim_type import parse_cpp_file, SimTypeFunction, SimTypeBottom
-from ...calling_conventions import DEFAULT_CC
-from ...misc import autoimport
-from ...misc.ux import once
-from ...sim_type import parse_file
-from ..stubs.ReturnUnconstrained import ReturnUnconstrained
-from ..stubs.syscall_stub import syscall as stub_syscall
+from angr.errors import AngrMissingTypeError
+from angr.sim_type import parse_cpp_file, SimTypeFunction, SimTypeBottom
+from angr.calling_conventions import DEFAULT_CC
+from angr.misc import autoimport
+from angr.misc.ux import once
+from angr.sim_type import parse_file
+from angr.procedures.stubs.ReturnUnconstrained import ReturnUnconstrained
+from angr.procedures.stubs.syscall_stub import syscall as stub_syscall
 
 if TYPE_CHECKING:
     from angr.calling_conventions import SimCCSyscall
@@ -25,8 +25,8 @@ if TYPE_CHECKING:
 
 
 l = logging.getLogger(name=__name__)
-SIM_LIBRARIES: Dict[str, "SimLibrary"] = {}
-SIM_TYPE_COLLECTIONS: Dict[str, "SimTypeCollection"] = {}
+SIM_LIBRARIES: dict[str, list[SimLibrary]] = {}
+SIM_TYPE_COLLECTIONS: dict[str, SimTypeCollection] = {}
 
 
 class SimTypeCollection:
@@ -35,15 +35,15 @@ class SimTypeCollection:
     """
 
     def __init__(self):
-        self.names: Optional[List[str]] = None
-        self.types: Dict[str, "SimType"] = {}
+        self.names: list[str] | None = None
+        self.types: dict[str, SimType] = {}
 
-    def set_names(self, *names):
-        self.names = names
+    def set_names(self, *names: str):
+        self.names = list(names)
         for name in names:
             SIM_TYPE_COLLECTIONS[name] = self
 
-    def add(self, name: str, t: "SimType") -> None:
+    def add(self, name: str, t: SimType) -> None:
         """
         Add a type to the collection.
 
@@ -53,7 +53,7 @@ class SimTypeCollection:
 
         self.types[name] = t
 
-    def get(self, name: str, bottom_on_missing: bool = False) -> "SimType":
+    def get(self, name: str, bottom_on_missing: bool = False) -> SimType:
         """
         Get a SimType object from the collection as identified by the name.
 
@@ -98,10 +98,10 @@ class SimLibrary:
     """
 
     def __init__(self):
-        self.type_collection_names: List[str] = []
+        self.type_collection_names: list[str] = []
         self.procedures = {}
         self.non_returning = set()
-        self.prototypes: Dict[str, SimTypeFunction] = {}
+        self.prototypes: dict[str, SimTypeFunction] = {}
         self.default_ccs = {}
         self.names = []
         self.fallback_cc = dict(DEFAULT_CC)
@@ -121,7 +121,7 @@ class SimLibrary:
         o.names = list(self.names)
         return o
 
-    def update(self, other):
+    def update(self, other: SimLibrary):
         """
         Augment this SimLibrary with the information from another SimLibrary
 
@@ -147,7 +147,10 @@ class SimLibrary:
         """
         for name in names:
             self.names.append(name)
-            SIM_LIBRARIES[name] = self
+            if name in SIM_LIBRARIES:
+                SIM_LIBRARIES[name].append(self)
+            else:
+                SIM_LIBRARIES[name] = [self]
 
     def set_default_cc(self, arch_name, cc_cls):
         """
@@ -252,7 +255,7 @@ class SimLibrary:
             proc.guessed_prototype = False
             if proc.prototype.arg_names is None:
                 # Use inspect to extract the parameters from the run python function
-                proc.prototype.arg_names = inspect.getfullargspec(proc.run).args[1:]
+                proc.prototype.arg_names = tuple(inspect.getfullargspec(proc.run).args[1:])
             if not proc.ARGS_MISMATCH:
                 proc.num_args = len(proc.prototype.args)
         if proc.display_name in self.non_returning:
@@ -273,8 +276,7 @@ class SimLibrary:
             proc = copy.deepcopy(self.procedures[name])
             self._apply_metadata(proc, arch)
             return proc
-        else:
-            return self.get_stub(name, arch)
+        return self.get_stub(name, arch)
 
     def get_stub(self, name, arch):
         """
@@ -292,7 +294,7 @@ class SimLibrary:
         self._apply_metadata(proc, arch)
         return proc
 
-    def get_prototype(self, name: str, arch=None) -> Optional[SimTypeFunction]:
+    def get_prototype(self, name: str, arch=None) -> SimTypeFunction | None:
         """
         Get a prototype of the given function name, optionally specialize the prototype to a given architecture.
 
@@ -345,17 +347,11 @@ class SimCppLibrary(SimLibrary):
 
     @staticmethod
     def _try_demangle(name):
-        if name[0:2] == "_Z":
-            try:
-                ast = itanium_demangler.parse(name)
-            except NotImplementedError:
-                return name
-            if ast:
-                return str(ast)
-        return name
+        ast = pydemumble.demangle(name)
+        return ast if ast else name
 
     @staticmethod
-    def _proto_from_demangled_name(name: str) -> Optional[SimTypeFunction]:
+    def _proto_from_demangled_name(name: str) -> SimTypeFunction | None:
         """
         Attempt to extract arguments and calling convention information for a C++ function whose name was mangled
         according to the Itanium C++ ABI symbol mangling language.
@@ -401,17 +397,16 @@ class SimCppLibrary(SimLibrary):
         stub = super().get_stub(demangled_name, arch)
         # try to determine a prototype from the function name if possible
         if demangled_name != name:
-            # itanium-mangled function name
+            # mangled function name
             stub.prototype = self._proto_from_demangled_name(demangled_name)
             if stub.prototype is not None:
                 stub.prototype = stub.prototype.with_arch(arch)
                 stub.guessed_prototype = False
                 if not stub.ARGS_MISMATCH:
-                    stub.cc.num_args = len(stub.prototype.args)
                     stub.num_args = len(stub.prototype.args)
         return stub
 
-    def get_prototype(self, name: str, arch=None) -> Optional[SimTypeFunction]:
+    def get_prototype(self, name: str, arch=None) -> SimTypeFunction | None:
         """
         Get a prototype of the given function name, optionally specialize the prototype to a given architecture. The
         function name will be demangled first.
@@ -468,10 +463,10 @@ class SimSyscallLibrary(SimLibrary):
 
     def __init__(self):
         super().__init__()
-        self.syscall_number_mapping: Dict[str, Dict[int, str]] = defaultdict(dict)  # keyed by abi
-        self.syscall_name_mapping: Dict[str, Dict[str, int]] = defaultdict(dict)  # keyed by abi
-        self.default_cc_mapping: Dict[str, Type["SimCCSyscall"]] = {}  # keyed by abi
-        self.syscall_prototypes: Dict[str, Dict[str, SimTypeFunction]] = defaultdict(dict)  # keyed by abi
+        self.syscall_number_mapping: dict[str, dict[int, str]] = defaultdict(dict)  # keyed by abi
+        self.syscall_name_mapping: dict[str, dict[str, int]] = defaultdict(dict)  # keyed by abi
+        self.default_cc_mapping: dict[str, type[SimCCSyscall]] = {}  # keyed by abi
+        self.syscall_prototypes: dict[str, dict[str, SimTypeFunction]] = defaultdict(dict)  # keyed by abi
         self.fallback_proc = stub_syscall
 
     def copy(self):
@@ -489,9 +484,10 @@ class SimSyscallLibrary(SimLibrary):
 
     def update(self, other):
         super().update(other)
-        self.syscall_number_mapping.update(other.syscall_number_mapping)
-        self.syscall_name_mapping.update(other.syscall_name_mapping)
-        self.default_cc_mapping.update(other.default_cc_mapping)
+        if isinstance(other, SimSyscallLibrary):
+            self.syscall_number_mapping.update(other.syscall_number_mapping)
+            self.syscall_name_mapping.update(other.syscall_name_mapping)
+            self.default_cc_mapping.update(other.default_cc_mapping)
 
     def minimum_syscall_number(self, abi):
         """
@@ -530,7 +526,7 @@ class SimSyscallLibrary(SimLibrary):
         :param mapping: A dict mapping syscall numbers to function names
         """
         self.syscall_number_mapping[abi].update(mapping)
-        self.syscall_name_mapping[abi].update(dict(reversed(i) for i in mapping.items()))
+        self.syscall_name_mapping[abi].update({b: a for a, b in mapping.items()})
 
     def set_abi_cc(self, abi, cc_cls):
         """
@@ -551,7 +547,7 @@ class SimSyscallLibrary(SimLibrary):
         """
         self.syscall_prototypes[abi][name] = proto
 
-    def set_prototypes(self, abi: str, protos: Dict[str, SimTypeFunction]) -> None:  # pylint: disable=arguments-differ
+    def set_prototypes(self, abi: str, protos: dict[str, SimTypeFunction]) -> None:  # pylint: disable=arguments-differ
         """
         Set the prototypes of many syscalls.
 
@@ -569,7 +565,7 @@ class SimSyscallLibrary(SimLibrary):
             mapping = self.syscall_number_mapping[abi]
             if number in mapping:
                 return mapping[number], arch, abi
-        return "sys_%d" % number, arch, None
+        return f"sys_{number}", arch, None
 
     def _apply_numerical_metadata(self, proc, number, arch, abi):
         proc.syscall_number = number
@@ -621,7 +617,7 @@ class SimSyscallLibrary(SimLibrary):
         l.debug("unsupported syscall: %s", number)
         return proc
 
-    def get_prototype(self, abi: str, name: str, arch=None) -> Optional[SimTypeFunction]:
+    def get_prototype(self, abi: str, name: str, arch=None) -> SimTypeFunction | None:
         """
         Get a prototype of the given syscall name and its ABI, optionally specialize the prototype to a given
         architecture.
@@ -686,7 +682,7 @@ class SimSyscallLibrary(SimLibrary):
 # - We will load all APIs when load_all_definitions() is called.
 
 _DEFINITIONS_BASEDIR = os.path.dirname(os.path.realpath(__file__))
-_EXTERNAL_DEFINITIONS_DIRS: Optional[List[str]] = None
+_EXTERNAL_DEFINITIONS_DIRS: list[str] | None = None
 
 
 def load_type_collections(skip=None) -> None:
@@ -743,8 +739,7 @@ def load_win32api_definitions():
         for _ in autoimport.auto_import_modules(
             "angr.procedures.definitions",
             _DEFINITIONS_BASEDIR,
-            filter_func=lambda module_name: module_name.startswith("win32_")
-            or module_name.startswith("wdk_")
+            filter_func=lambda module_name: module_name.startswith(("win32_", "wdk_"))
             or module_name in {"ntoskrnl", "ntdll", "user32"},
         ):
             pass
