@@ -1,25 +1,11 @@
 # pylint:disable=line-too-long
+from __future__ import annotations
 import logging
 from collections import OrderedDict
 
-from ...sim_type import (SimTypeFunction,
-    SimTypeShort,
-    SimTypeInt,
-    SimTypeLong,
-    SimTypeLongLong,
-    SimTypeDouble,
-    SimTypeFloat,
-    SimTypePointer,
-    SimTypeChar,
-    SimStruct,
-    SimTypeArray,
-    SimTypeBottom,
-    SimUnion,
-    SimTypeBool,
-    SimTypeRef,
-)
-from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
-from .. import SIM_PROCEDURES as P
+from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
+from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
+from angr.procedures import SIM_PROCEDURES as P
 from . import SimLibrary
 
 
